import json
import os
import time
from datetime import datetime

import click
from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.config import Config
from cnvrgv2.config import CONFIG_FOLDER_NAME
from cnvrgv2.modules.dataset import Dataset


@click.group(name='dataset')
def dataset_group():
    pass


@dataset_group.command()
@click.option('-n', '--name', prompt=messages.DATASET_PROMPT_CLONE, help=messages.DATASET_HELP_CLONE)
@click.option('-o', '--override', is_flag=True, default=False, help=messages.DATASET_HELP_CLONE_OVERRIDE)
@click.option('-cl', '--cache_link', default=False, is_flag=True, required=False,
              help=messages.DATASET_HELP_CLONE_OVERRIDE)
@click.option('-c', '--commit', default=None, help=messages.DATASET_HELP_CLONE_COMMIT)
@click.option('-t', '--threads', default=40, type=int, help=messages.CLONE_NUMBER_OF_THREADS)
@click.option('-q', '--query', default=None, help=messages.QUERY_SLUG)
@prepare_command()
def clone(cnvrg, logger, name, override, cache_link, commit, threads, query):
    """
    Clones the given dataset to local folder
    """
    dataset = cnvrg.datasets.get(name)
    logger.info(messages.LOG_CLONING_DATASET.format(name))
    if os.path.exists(dataset.slug + '/' + CONFIG_FOLDER_NAME) and not override:
        logger.log_and_echo(messages.DATASET_CLONE_SKIP.format(name))
        return
    dataset.clone(progress_bar_enabled=True, override=override, commit=commit, use_cached=cache_link, threads=threads,
                  query_slug=query)
    success_message = messages.DATASET_CLONE_SUCCESS.format(name)
    logger.log_and_echo(success_message)


@dataset_group.command()
@click.option('-n', '--name', prompt=messages.DATASET_PROMPT_NAME, help=messages.DATASET_HELP_NAME)
@click.option('-f', '--files', prompt=messages.DATASET_PUT_PROMPT_FILES, help=messages.DATASET_PUT_HELP_FILES)
@click.option('-fc', '--force', is_flag=True, help=messages.DATA_UPLOAD_HELP_FORCE)
@click.option('-or', '--override', is_flag=True, help=messages.DATA_UPLOAD_HELP_OVERRIDE)
@click.option('-gd', '--git-diff', is_flag=True, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@prepare_command()
def put(cnvrg, logger, name, files, force, override, git_diff):
    """
    Uploads the given files to the given dataset
    """
    file_paths = files.split(",")
    dataset = cnvrg.datasets.get(name)
    dataset.put_files(paths=file_paths, progress_bar_enabled=True, git_diff=git_diff, force=force, override=override)
    logger.log_and_echo(messages.DATA_UPLOAD_SUCCESS)


@dataset_group.command()
@click.option('-g', '--git-diff', is_flag=True, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@prepare_command()
def upload(dataset, logger, git_diff):
    """
    Uploads updated files from the current dataset folder
    """
    dataset.upload(progress_bar_enabled=True, git_diff=git_diff)
    logger.log_and_echo(messages.DATA_UPLOAD_SUCCESS)


@dataset_group.command()
@click.option('-c', '--commit', default=None, help=messages.PROJECT_DOWNLOAD_HELP_COMMIT)
@prepare_command()
def download(dataset, logger, commit):
    """
    Downloads updated files to the current dataset folder
    """
    dataset.download(progress_bar_enabled=True, commit_sha1=commit)
    logger.log_and_echo(messages.DATA_DOWNLOAD_SUCCESS)


@dataset_group.command()
@click.option('-n', '--name', prompt=messages.DATASET_PROMPT_NAME, help=messages.DATASET_HELP_CLONE)
@click.option('-f', '--files', prompt=messages.DATASET_REMOVE_PROMPT_FILES, help=messages.DATASET_REMOVE_HELP_FILES)
@click.option('-m', '--message', help=messages.DATA_COMMIT_MESSAGE, default="")
@prepare_command()
def remove(cnvrg, logger, name, files, message):
    """
    Removes the given files remotely
    """
    file_paths = files.split(",")
    dataset = cnvrg.datasets.get(name)
    dataset.remove_files(paths=file_paths, message=message, progress_bar_enabled=True)
    logger.log_and_echo(messages.DATASET_REMOVE_SUCCESS)


@dataset_group.command()
@click.option('-n', '--name', prompt=messages.DATASET_PROMPT_DELETE, help=messages.DATASET_HELP_DELETE)
@prepare_command()
def delete(cnvrg, logger, name):
    """
    Deletes the dataset
    """
    click.confirm(messages.DATASET_DELETE_CONFIRM.format(name), default=False, abort=True)
    dataset = cnvrg.datasets.get(name)
    dataset.delete()
    success_message = messages.DATASET_DELETE_SUCCESS.format(name)
    logger.log_and_echo(success_message)


@dataset_group.command()
@click.option('-j', '--job-slug', default=None, help=messages.PROJECT_JOB_SLUG_HELP_NAME)
@click.option('-g', '--git-diff', is_flag=True, default=None, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@click.option('-m', '--message', default="", help=messages.DATA_COMMIT_MESSAGE)
@prepare_command()
def sync(dataset, logger, message, job_slug=None, git_diff=None):
    """
    Sync local project to remote
    """
    logger.log_and_echo('Syncing...')
    dataset.sync(job_slug=job_slug, git_diff=git_diff, progress_bar_enabled=True, message=message)
    logger.log_and_echo(messages.DATASET_SYNC_SUCCESS.format(dataset.slug))


@dataset_group.command()
@click.option('-n', '--name', prompt=messages.DATASET_PROMPT_CREATE, help=messages.DATASET_HELP_CREATE)
@prepare_command()
def create(cnvrg, logger, name):
    """
    Creates a new dataset, and associates the current folder with the created project.
    It is recommended to create a new dataset in an empty folder.
    To create a dataset from a folder that contains content please refer to dataset link command.
    """
    if os.listdir():
        click.confirm(messages.DATASET_CREATE_FOLDER_NOT_EMPTY, default=False, abort=True)

    logger.log_and_echo(messages.DATASET_CREATING_MESSAGE.format(name))
    new_dataset = cnvrg.datasets.create(name)

    new_dataset.local_commit = new_dataset.last_commit

    logger.log_and_echo(messages.DATASET_CONFIGURING_FOLDER)
    new_dataset.save_config()
    logger.log_and_echo(messages.DATASET_CREATE_SUCCESS.format(new_dataset.title))


@dataset_group.command(name='list')
@prepare_command()
def list_command(cnvrg, logger):
    """
    list all datasets
    """
    datasets = cnvrg.datasets.list()
    domain = cnvrg._context.domain
    organization = cnvrg._context.organization
    table = PrettyTable()
    table.field_names = ["Dataset Title", "Dataset Link", "Public"]
    table.align["Dataset Title"] = "l"
    table.align["Dataset Link"] = "l"
    table.align["Public"] = "l"
    for dataset in datasets:
        url = "{0}/{1}/datasets/{2}".format(domain, organization, dataset.slug)
        table.add_row([dataset.title, url, dataset.is_public])
    click.echo(table)


@dataset_group.command()
@click.option('-n', '--name', prompt=messages.DATASET_PROMPT_NAME, help=messages.DATASET_HELP_CACHE)
@click.option('-c', '--commit', prompt=messages.DATA_PROMPT_CACHE, default="latest", show_default=True,
              help=messages.DATASET_HELP_CACHE)
@click.option('-d', '--external-disk-title', prompt=messages.DATASET_PROMPT_DISK_NAME, help=messages.DATASET_HELP_CACHE)
@prepare_command()
def cache(cnvrg, logger, name, commit, external_disk_title):
    """
    Caches the current commit onto an external disk, for quick access
    """
    dataset = cnvrg.datasets.get(name)
    if commit.lower() == 'latest':
        commit = dataset.last_commit
    dataset_commit = dataset.get_commit(commit)
    dataset_commit.cache_commit(external_disk_title)
    logger.log_and_echo(messages.DATASET_CACHE_SUCCESS)


@dataset_group.command()
@click.option('-n', '--name', prompt=messages.DATASET_PROMPT_NAME, help=messages.DATASET_HELP_UNCACHE)
@click.option('-c', '--commit', prompt=messages.DATA_PROMPT_CACHE, default="latest", show_default=True,
              help=messages.DATASET_HELP_UNCACHE)
@click.option('-d', '--external-disk-title', prompt=messages.DATASET_PROMPT_DISK_NAME,
              help=messages.DATASET_HELP_UNCACHE)
@prepare_command()
def uncache(cnvrg, logger, name, commit, external_disk_title):
    """
    Removes the current cached commit from the external disk.
    """
    dataset = cnvrg.datasets.get(name)
    if commit.lower() == 'latest':
        commit = dataset.last_commit
    dataset_commit = dataset.get_commit(commit)
    dataset_commit.clear_cached_commit(external_disk_title)
    logger.log_and_echo(messages.DATASET_UNCACHE_SUCCESS)


@dataset_group.command()
@prepare_command()
def scan(logger):
    """
    Scan the current working directory and print the datasets found in it
    """
    logger.log_and_echo(messages.DATASET_SCAN_START)
    cwd = os.getcwd()
    try:
        datasets = []
        _, dirs, _ = next(os.walk(cwd))  # Direct children of cwd

        for curr_dir in dirs:
            os.chdir(os.path.join(cwd, curr_dir))
            config = Config()
            if config.local_config:
                datasets.append({
                    "dataset_slug": config.dataset_slug,
                    "local_commit": config.commit_sha1
                })

        if len(datasets) > 0:
            logger.log_and_echo(json.dumps(datasets))
        else:
            logger.log_and_echo(messages.DATASET_SCAN_NO_RESULTS)
    finally:
        os.chdir(cwd)


@dataset_group.command()
@click.option('-n', '--names', prompt=messages.DATASET_NAMES, help=messages.DATASET_NAMES_HELP)
@click.option('-t', '--time-out', help=messages.DATASET_VERIFY_TIMEOUT,
              required=False, default=None)
@prepare_command()
def verify(cnvrg, logger, names, time_out):
    """
    Scans datasets folders for given names and returns when all of them are successfully downloaded,
    returns status code 1 otherwise (in use for Agent)
    """
    datasets = names.split(",")
    start_time = datetime.now()
    all_ready = [False] * len(datasets)
    max_timeout = int(time_out) if time_out and time_out.isnumeric() else None

    while True:
        try:
            seconds_diff = (datetime.now() - start_time).total_seconds()
            if max_timeout:
                if seconds_diff >= max_timeout:
                    logger.log_and_echo(messages.DATASET_VERIFY_STATUS.format(all(all_ready)),
                                        error=False if all(all_ready) else True)
                    return
                else:
                    # For debugging purposes
                    logger.log_and_echo(str('%.1f' % (max_timeout - seconds_diff)) + " seconds until timeout")
            for i, dataset in enumerate(datasets):
                # For debugging purposes
                logger.log_and_echo("Verifying " + dataset)
                if Dataset(slug=dataset).verify():
                    all_ready[i] = True
            if all(all_ready):
                logger.log_and_echo(messages.DATASET_VERIFY_STATUS.format(True))
                return
            else:
                time.sleep(1)
        except FileNotFoundError:
            continue
    # return exit status 1 by passing error=True(in use for Agent)
    logger.log_and_echo(messages.DATASET_VERIFY_STATUS.format(False), error=True)
    return



from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.data.local_files_handler import LocalFilesHandler


class FileUploader(LocalFilesHandler):
    def __init__(
        self,
        data_owner,
        paths,
        commit,
        metadata,
        override=False,
        num_workers=40,
        chunk_size=1000,
        queue_size=5000,
        progress_bar_enabled=False
    ):
        """
        Multithreaded local file uploader
        @param data_owner: Cnvrg dataset / project object
        @param paths: [Generator] that lists paths to upload
        @param commit: The commit sha1 to which we upload the files
        @param metadata: [Dict] The metadata on the uploaded files (total files, file size..)
        @param override: Boolean stating whether or not we should re-upload even if the file already exists
        @param num_workers: Number of threads to use for concurrent file uploading
        @param chunk_size: File meta chunk size to fetch from the server
        @param queue_size: Max number of file meta to put in queue
        @param progress_bar_enabled: Boolean indicating whenever or not to print a progress bar. In use of the cli
        """
        self.commit = commit
        super().__init__(
            data_owner,
            paths=paths,
            metadata=metadata,
            override=override,
            num_workers=num_workers,
            chunk_size=chunk_size,
            queue_size=queue_size,
            progress_bar_enabled=progress_bar_enabled
        )

    def _handle_file_progress_function(self, uploaded_ids):
        """
        Function to progress the upload
        @param uploaded_ids: Uploaded file ids
        @return: None
        """
        data = {"file_ids": uploaded_ids}
        self.data_owner._proxy.call_api(
            route=urljoin(self.data_owner._route, "commits", self.commit, "upload_files_save"),
            http_method=HTTP.POST,
            payload=data
        )

    def _file_handler_function(self, local_path, object_path, progress_bar=None):
        """
        Function to upload single file
        @param local_path: File location locally
        @param object_path: File location in bucket
        @return: None
        """
        self.storage_client.upload_single_file(local_path, object_path, progress_bar)

    def _file_collector_function(self, files_metadata):
        """
        Function to collect files metadata from server
        @param files_metadata: Local files metadata to sent to the server for getting files to upload
        @return: Array of files metadata
        """
        data = {
            "files": files_metadata,
            "override": self.override,
        }
        response = self.data_owner._proxy.call_api(
            route=urljoin(self.data_owner._route, "commits", self.commit, "upload_files"),
            http_method=HTTP.POST,
            payload=data
        )

        blob_array = []
        for blob in response.items:
            blob_array.append({"id": blob.id, **blob.attributes})

        return blob_array



from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.config import routes, error_messages
from cnvrgv2.utils.validators import validate_email, validate_user_role
from cnvrgv2 import cnvrg
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.members import Member
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config.error_messages import FAULTY_VALUE


class MembersClient:
    def __init__(self, organization):
        if type(organization) is not cnvrg.Cnvrg:
            raise CnvrgArgumentsError(error_messages.ARGUMENT_BAD_TYPE.format(cnvrg.Cnvrg, type(organization)))

        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._scope = scope
        self._proxy = Proxy(context=self._context)
        self._list_route = routes.MEMBERS_ALL_BASE.format(scope["organization"])
        self._route = routes.MEMBERS_BASE.format(scope["organization"])

    def add(self, email, role):
        """
        Add a new member to the organization
        @param email: User email
        @param role: Role to add
        """

        if not validate_user_role(role):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(role))

        if not validate_email(email):
            raise CnvrgArgumentsError(FAULTY_VALUE.format(email))

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="users", attributes={"members": [{"email": email, "role": role}]})
        )

    def get(self, email):
        """
        Retrieves a project by the given slug
        @param email: The slug of the requested project
        @return: Project object
        """
        if not email or not isinstance(email, str):
            raise CnvrgArgumentsError(error_messages.INVALID_EMAIL)

        return Member(context=self._context, email=email)

    def list(self, filter=None, sort="-id"):
        """
        List all members in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @param filter: Filter json
        @raise: HttpError
        @return: Generator that yields dataset objects
        """

        context = self._context

        def member_builder(item):
            return Member(context, attributes=item.attributes)

        return api_list_generator(
            context=self._context,
            route=self._list_route,
            object_builder=member_builder,
            identifier="username",
            sort=sort,
            filter=filter
        )




import json

from cnvrgv2.proxy import Proxy, HTTP


def api_list_generator(
    context,
    route,
    object=None,
    object_builder=None,
    filter=None,
    sort="-id",
    page_size=20,
    identifier="slug",
    pagination_type="cursor",
    data=None
):
    """
    General api list generator for smart resource paginating
    @param context: Proxy object to communicate with the API
    @param route: The relevant API route
    @param object: The object to cast the response into
    @param object_builder: Function to build a specific object
    @param sort: sort key (-key -> DESC | key -> ASC)
    @param page_size: Number of items returned each API call
    @param identifier: Name of the object's identifier, in case it's not slug
    @param pagination_type: Cursor or Offset pagination
    @param data: Additional data to add to the list call
    @return: Generator object
    """

    proxy = Proxy(context)
    next_page = ""

    while True:
        data = data or {}

        payload = {"sort": sort, "page[size]": page_size, **data}

        # Using dumps on None creates the string "null" which is unexpected at server
        if filter:
            payload["filter"] = json.dumps(filter)

        if pagination_type == "cursor":
            payload["page[after]"] = next_page
        else:
            if next_page == "":
                next_page = 1
            payload["page[offset]"] = next_page

        response = proxy.call_api(
            route=route,
            http_method=HTTP.GET,
            payload=payload
        )
        items = response.items

        for item in items:
            object_slug = item.attributes[identifier]
            if object_builder:
                object_instance = object_builder(item)
            else:
                object_instance = object(context=context, slug=object_slug, attributes=item.attributes)
            yield object_instance

        if len(items) < page_size:
            break

        next_page = response.next if pagination_type == "cursor" else next_page + 1



import io
import re
import time
from datetime import datetime
import pytz
import os

# Log levels
LOGS_TYPE_OUTPUT = "output"
LOGS_TYPE_ERROR = "error"
LOGS_TYPE_INFO = "info"
LOGS_TYPE_WARNING = "warning"

# Log filter types
FILTER_LOGS_STDOUT = "stdout"
FILTER_LOGS_CNVRG_ERROR = "cnvrg-error"
FILTER_LOGS_CNVRG_INFO = "cnvrg-info"


MAX_LOGS_PER_SEND = int(os.environ.get("CNVRG_MAX_LOGS_PER_SEND") or 500)

ANSI_ESCAPE = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')


def timestamp_for_logs():
    """
    @return: A string representing current UTC time
    """
    return datetime.utcnow().replace(tzinfo=pytz.utc).strftime("%Y-%m-%dT%H:%M:%S%z")


def extract_lines(block):
    """
    Extracts logs from a log buffer
    @param block: String of buffered logs, from stdout/stderr
    @return: List of parsed logs
    """
    return [ANSI_ESCAPE.sub('', line) for line in block.strip().split("\n")]


def log_buffer(workflow, log_type, buffer):
    """
    Function that parses a buffer holding logs and sends them to the corresponding workflow at the server
    The function will run until the buffer is closed or eof is present.
    @param workflow: Current workflow's object. Used to send logs to the server
    @param log_type: The log level to be written
    @param buffer: The buffer to be parsed (StringIo or BufferedReader)
    @return: None
    """
    if isinstance(buffer, io.StringIO):
        seek = 0
        while True:
            buffer_value = buffer.getvalue()
            lines = extract_lines(buffer_value[seek:].strip('\0'))
            workflow.write_logs(lines, log_type=log_type)
            seek = len(buffer_value)

            if '\0' in buffer_value:
                break
            time.sleep(0.5)
    else:
        # Is BufferedReader
        for log in buffer:
            log = log.decode("utf-8")
            workflow.write_logs(log, log_type=log_type)
